"""Flow class."""
from typing import Any, Dict, Sequence

from uniflow.constants import TRANSFORM
from uniflow.flow.flow import Flow
from uniflow.node import Node
from uniflow.op.basic.copy_op import CopyOp
from uniflow.op.prompt_schema import GuidedPrompt


class TransformCopyFlow(Flow):
    """Copy flow class.

    This is a demo flow does nothing but copy the input nodes.
    """

    TAG = TRANSFORM

    def __init__(
        self,
        guided_prompt_template: GuidedPrompt,
        model_config: Dict[str, Any],
    ) -> None:  # pylint: disable=useless-parent-delegation
        """Initialize CopyFlow class."""
        self._copy_op = CopyOp(name="copy_op")
        super().__init__()

    def run(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run CopyFlow.

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        return self._copy_op(nodes)



"""LLM rater for tasks with a LLM model server."""

import re
from collections import Counter
from typing import Any, Dict, List

from uniflow.op.model.constants import (
    AVERAGE_SCORE,
    MAJORITY_VOTE,
    RESPONSE,
    SCORES,
    VOTES,
)
from uniflow.op.model.llm_processor import JsonFormattedDataProcessor, LLMDataProcessor
from uniflow.op.prompt_schema import GuidedPrompt


class LLMRater(LLMDataProcessor):
    """LLM Rater"""

    def __init__(
        self,
        guided_prompt_template: GuidedPrompt,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """LLM Rater Constructor.

        Args:
            guided_prompt_template (GuidedPrompt): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__(guided_prompt_template, model_config)
        example_keys = list(guided_prompt_template.examples[0].dict().keys())
        pattern = r"^[^A-Za-z]+|[^A-Za-z]+$"
        self._rater_key = example_keys[-1]
        self._label2score = {
            re.sub(pattern, "", k).lower().lower(): float(v)
            for k, v in label2score.items()
        }
        self._score2label = {v: k for k, v in self._label2score.items()}

    def _deserialize(self, data: List[str]) -> List[Dict[str, Any]]:
        """Deserialize data.

        Args:
            data (List[str]): Data to deserialize.

        Returns:
            List[Dict[str, Any]]: Deserialized data.
        """

        def _extract_label(text):
            # Create a regular expression pattern from the dictionary keys
            # The pattern will look like '(yes|no)', case-insensitive
            pattern = (
                r"\b("
                + "|".join(  # noqa: W503
                    re.escape(key) for key in self._label2score.keys()
                )  # noqa: W503
                + r")\b"  # noqa: W503
            )

            matches = list(re.finditer(pattern, text, re.IGNORECASE))
            if matches:
                # Return the last matched group (Yes or No), preserving the original case
                return (
                    matches[-1].group(1).lower()
                )  # Convert to lowercase to match dictionary keys
            else:
                # Return None if no match is found
                return None

        data = super()._deserialize(data)
        response = data[RESPONSE]

        labels = [_extract_label(d) for d in response]
        scores = []
        for label in labels:
            if label is not None:
                scores.append(self._label2score[label])
        majority_vote = Counter(labels).most_common(1)[0][0]
        mean_score = sum(scores) / len(scores) if len(scores) > 0 else None

        data.update(
            {
                MAJORITY_VOTE: majority_vote,
                AVERAGE_SCORE: mean_score,
                VOTES: labels,
                SCORES: scores,
            }
        )

        return data


class JsonFormattedLLMRater(JsonFormattedDataProcessor):
    """Json Formatted LLM Rater"""

    def __init__(
        self,
        guided_prompt_template: GuidedPrompt,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """Json Formatted LLM Rater Constructor.

        Args:
            guided_prompt_template (GuidedPrompt): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__(guided_prompt_template, model_config)
        example_keys = list(guided_prompt_template.examples[0].dict().keys())
        self._pattern = r"^[^A-Za-z]+|[^A-Za-z]+$"
        self._rater_key = example_keys[-1]
        self._label2score = {
            re.sub(self._pattern, "", k).lower(): float(v)
            for k, v in label2score.items()
        }
        self._score2label = {v: k for k, v in self._label2score.items()}

    def _deserialize(self, data: List[str]) -> List[Dict[str, Any]]:
        """Deserialize data.

        Args:
            data (List[str]): Data to deserialize.

        Returns:
            List[Dict[str, Any]]: Deserialized data.
        """
        data = super()._deserialize(data)
        response = data[RESPONSE]

        labels = [
            re.sub(self._pattern, "", r[self._rater_key]).lower()
            if self._rater_key in r
            else None
            for r in response
        ]
        scores = []
        for label in labels:
            if label is not None and label in self._label2score:
                scores.append(self._label2score[label])
        majority_vote = Counter(labels).most_common(1)[0][0]
        mean_score = sum(scores) / len(scores) if len(scores) > 0 else None
        data.update(
            {
                MAJORITY_VOTE: majority_vote,
                AVERAGE_SCORE: mean_score,
                VOTES: labels,
                SCORES: scores,
            }
        )

        return data



"""
All Model Servers including ModelServerFactory, AbsModelServer, OpenAIModelServer and HuggingfaceModelServer.
"""

import re
from functools import partial
from typing import Any, Dict, List

from uniflow.op.model.model_config import (
    AzureOpenAIModelConfig,
    HuggingfaceModelConfig,
    LMQGModelConfig,
    NougatModelConfig,
    OpenAIModelConfig,
)

###############################################################################
#                             All Model Servers                               #
###############################################################################


class ModelServerFactory:
    """Model Server Factory."""

    _servers = {}

    @classmethod
    def register(cls, name: str, server_cls: "AbsModelServer") -> None:
        """Register model server.

        Args:
            name (str): Model server name.
            server_cls (AbsModelServer): Model server class.
        """
        cls._servers[name] = server_cls

    @classmethod
    def get(cls, name: str) -> "AbsModelServer":
        """Get model server.

        Args:
            name (str): Model server name.

        Returns:
            AbsModelServer: Model server.

        Raises:
            ValueError: If no model server registered under the name.
        """
        server_cls = cls._servers.get(name)
        if not server_cls:
            raise ValueError(f"No model server registered under '{name}'")
        return server_cls

    @classmethod
    def list(cls):
        """List all registered model servers.

        Returns:
            List[str]: List of registered model server names.
        """
        return list(cls._servers.keys())


class AbsModelServer:
    """Abstract Model Server Class."""

    def __init_subclass__(cls):
        """Register model server.

        This method is called when subclass is created.
        """
        super().__init_subclass__()
        ModelServerFactory.register(cls.__name__, cls)

    def __init__(self, model_config: Dict[str, Any]) -> None:
        """Initialize AbsModelServer class.

        Args:
            model_config (Dict[str, Any]): Model config.
        """
        self._model_config = model_config

    def _preprocess(self, data: str) -> str:
        """Preprocess data.

        Args:
            data (str): Data to preprocess.

        Returns:
            str: Preprocessed data.
        """
        raise NotImplementedError

    def __call__(self, data: str) -> str:
        """Run model.

        Args:
            data (str): Data to run.

        Returns:
            str: Output data.
        """
        raise NotImplementedError

    def _postprocess(self, data: str) -> List[str]:
        """Postprocess data.

        Args:
            data (str): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        raise NotImplementedError


class OpenAIModelServer(AbsModelServer):
    """OpenAI Model Server Class."""

    def __init__(self, model_config: Dict[str, Any]) -> None:
        # import in class level to avoid installing openai package
        from openai import OpenAI  # pylint: disable=import-outside-toplevel

        super().__init__(model_config)
        self._model_config = OpenAIModelConfig(**self._model_config)
        self._client = OpenAI()

    def _preprocess(self, data: List[str]) -> List[str]:
        """Preprocess data.

        Args:
            data (List[str]): Data to preprocess.

        Returns:
            List[str]: Preprocessed data.
        """
        return data

    def _postprocess(self, data: List[str]) -> List[str]:
        """Postprocess data.

        Args:
            data (str): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        return [c.message.content for d in data for c in d.choices]

    def __call__(self, data: List[str]) -> List[str]:
        """Run model.

        OpenAI completions API does not support batch inference.

        Args:
            data (str): Data to run.

        Returns:
            str: Output data.
        """
        data = self._preprocess(data)
        inference_data = []
        for d in data:
            inference_data.append(
                self._client.chat.completions.create(
                    model=self._model_config.model_name,
                    messages=[
                        {"role": "user", "content": d},
                    ],
                    n=self._model_config.num_call,
                    temperature=self._model_config.temperature,
                    response_format=self._model_config.response_format,
                )
            )
        data = self._postprocess(inference_data)
        return data


class AzureOpenAIModelServer(AbsModelServer):
    """Azure OpenAI Model Server Class."""

    def __init__(self, model_config: Dict[str, Any]) -> None:
        # import in class level to avoid installing openai package
        from openai import AzureOpenAI  # pylint: disable=import-outside-toplevel

        super().__init__(model_config)
        self._model_config = AzureOpenAIModelConfig(**self._model_config)
        self._client = AzureOpenAI(
            api_key=self._model_config.api_key,
            api_version=self._model_config.api_version,
            azure_endpoint=self._model_config.azure_endpoint,
        )

    def _preprocess(self, data: List[str]) -> List[str]:
        """Preprocess data.

        Args:
            data (List[str]): Data to preprocess.

        Returns:
            List[str]: Preprocessed data.
        """
        return data

    def _postprocess(self, data: List[str]) -> List[str]:
        """Postprocess data.

        Args:
            data (str): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        return [c.message.content for d in data for c in d.choices]

    def __call__(self, data: List[str]) -> List[str]:
        """Run model.

        Azure OpenAI completions API does not support batch inference.

        Args:
            data (str): Data to run.

        Returns:
            str: Output data.
        """
        data = self._preprocess(data)
        inference_data = []
        for d in data:
            inference_data.append(
                self._client.chat.completions.create(
                    model=self._model_config.model_name,
                    messages=[
                        {"role": "user", "content": d},
                    ],
                    n=self._model_config.num_call,
                    temperature=self._model_config.temperature,
                    response_format=self._model_config.response_format,
                )
            )
        data = self._postprocess(inference_data)
        return data


class HuggingfaceModelServer(AbsModelServer):
    """Huggingface Model Server Class."""

    def __init__(self, model_config: Dict[str, Any]) -> None:
        # import in class level to avoid installing transformers package
        from transformers import pipeline  # pylint: disable=import-outside-toplevel
        from transformers import (  # pylint: disable=import-outside-toplevel
            AutoModelForCausalLM,
            AutoTokenizer,
        )

        super().__init__(model_config)
        self._model_config = HuggingfaceModelConfig(**self._model_config)

        tokenizer = AutoTokenizer.from_pretrained(
            self._model_config.model_name,
        )
        tokenizer.pad_token = tokenizer.eos_token

        model = AutoModelForCausalLM.from_pretrained(
            self._model_config.model_name,
            device_map="auto",
            offload_folder="./offload",
            load_in_4bit=True,
        )

        # explicitly set batch_size for pipeline
        # for batch inference.
        self._pipeline = pipeline(
            "text-generation",
            model=model,
            tokenizer=tokenizer,
            device_map="auto",
            max_new_tokens=768,
            num_return_sequences=1,
            repetition_penalty=1.2,
            eos_token_id=tokenizer.eos_token_id,
            pad_token_id=tokenizer.pad_token_id,
            batch_size=self._model_config.batch_size,
        )

    def _preprocess(self, data: List[str]) -> List[str]:
        """Preprocess data.

        Args:
            data (List[str]): Data to preprocess.

        Returns:
            List[str]: Preprocessed data.
        """
        return data

    def _postprocess(self, data: List[str]) -> List[str]:
        """Postprocess data.

        Args:
            data (List[str]): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        return [d["generated_text"] for output_list in data for d in output_list]

    def __call__(self, data: List[str]) -> List[str]:
        """Run model.

        Args:
            data (List[str]): Data to run.

        Returns:
            List[str]: Output data.
        """
        data = self._preprocess(data)
        data = self._pipeline(data)
        data = self._postprocess(data)
        return data


class LMQGModelServer(AbsModelServer):
    """Huggingface Model Server Class."""

    def __init__(self, model_config: Dict[str, Any]) -> None:
        # import in class level to avoid installing transformers package
        from lmqg import TransformersQG  # pylint: disable=import-outside-toplevel

        super().__init__(model_config)
        self._model_config = LMQGModelConfig(**self._model_config)

        self._model = TransformersQG(
            model=self._model_config.model_name, max_length=1024
        )

    def _preprocess(self, data: List[str]) -> List[str]:
        """Preprocess data.

        Args:
            data (List[str]): Data to preprocess.

        Returns:
            List[str]: Preprocessed data.
        """
        return data

    def _postprocess(self, data: List[str]) -> List[str]:
        """Postprocess data.

        Args:
            data (List[str]): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        return data

    def __call__(self, data: List[str]) -> List[str]:
        """Run model.

        Args:
            data (List[str]): Data to run.

        Returns:
            List[str]: Output data.
        """
        data = self._preprocess(data)
        data = self._model.generate_qa(data)
        data = self._postprocess(data)
        return data


class NougatModelServer(AbsModelServer):
    """Nougat Model Server Class."""

    def __init__(self, model_config: Dict[str, Any]) -> None:
        # import in class level to avoid installing nougat package
        try:
            from nougat import NougatModel  # pylint: disable=import-outside-toplevel
            from nougat.utils.checkpoint import (
                get_checkpoint,  # pylint: disable=import-outside-toplevel
            )
            from nougat.utils.device import (
                move_to_device,  # pylint: disable=import-outside-toplevel
            )
        except ModuleNotFoundError as exc:
            raise ModuleNotFoundError(
                "Please install nougat to use NougatModelServer. You can use `pip install nougat-ocr` to install it."
            ) from exc

        super().__init__(model_config)
        self._model_config = NougatModelConfig(**self._model_config)
        checkpoint = get_checkpoint(None, model_tag=self._model_config.model_name)
        self.model = NougatModel.from_pretrained(checkpoint)
        self.model = move_to_device(
            self.model, bf16=False, cuda=self._model_config.batch_size > 0
        )
        self.model.eval()

    def _preprocess(self, data: str) -> List[str]:
        """Preprocess data.

        Args:
            data (List[str]): Data to preprocess.

        Returns:
            List[str]: Preprocessed data.
        """
        return data

    def _postprocess(self, data: List[str]) -> List[str]:
        """Postprocess data.

        Args:
            data (List[str]): Data to postprocess.

        Returns:
            List[str]: Postprocessed data.
        """
        return [d["generated_text"] for output_list in data for d in output_list]

    def __call__(self, data: List[str]) -> List[str]:
        """Run model.

        Args:
            data (List[str]): Data to run.

        Returns:
            List[str]: Output data.
        """
        from nougat.postprocessing import (
            markdown_compatible,  # pylint: disable=import-outside-toplevel
        )
        from nougat.utils.dataset import (
            LazyDataset,  # pylint: disable=import-outside-toplevel
        )
        from torch.utils.data import (  # pylint: disable=import-outside-toplevel
            ConcatDataset,
            DataLoader,
        )

        outs = []
        for pdf in data:
            dataset = LazyDataset(
                pdf,
                partial(self.model.encoder.prepare_input, random_padding=False),
                None,
            )
            dataloader = DataLoader(
                ConcatDataset([dataset]),
                batch_size=1,
                shuffle=False,
                collate_fn=LazyDataset.ignore_none_collate,
            )
            predictions = []
            page_num = 0
            for i, (sample, is_last_page) in enumerate(dataloader):
                model_output = self.model.inference(
                    image_tensors=sample, early_stopping=False
                )
                # check if model output is faulty
                for j, output in enumerate(model_output["predictions"]):
                    page_num += 1
                    if output.strip() == "[MISSING_PAGE_POST]":
                        # uncaught repetitions -- most likely empty page
                        predictions.append(f"\n\n[MISSING_PAGE_EMPTY:{page_num}]\n\n")
                    else:
                        output = markdown_compatible(output)
                        predictions.append(output)
                    if is_last_page[j]:
                        out = "".join(predictions).strip()
                        out = re.sub(r"\n{3,}", "\n\n", out).strip()
            outs.append(out)
        return outs



"""Classify flow for single point auto-rater."""

from typing import Any, Dict, Sequence

from uniflow.constants import RATER
from uniflow.flow.flow import Flow
from uniflow.node import Node
from uniflow.op.model.llm_rater import JsonFormattedLLMRater, LLMRater
from uniflow.op.model.model_op import ModelOp
from uniflow.op.prompt_schema import GuidedPrompt


class RaterFlow(Flow):
    """Rater Classification Flow."""

    TAG = RATER

    def __init__(
        self,
        guided_prompt_template: GuidedPrompt,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """Rater Flow Constructor.

        Args:
            guided_prompt_template (GuidedPrompt): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__()
        if (
            "response_format" in model_config
            and model_config["response_format"]["type"] == "json_object"  # noqa: W503
        ):
            model = JsonFormattedLLMRater(
                guided_prompt_template=guided_prompt_template,
                model_config=model_config,
                label2score=label2score,
            )
        else:
            model = LLMRater(
                guided_prompt_template=guided_prompt_template,
                model_config=model_config,
                label2score=label2score,
            )
        self._model_op = ModelOp(
            name="rater_model_op",
            model=model,
        )

    def run(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run Model Flow.

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        return self._model_op(nodes)



"""Extract, Transform and Pipeline config module."""

from dataclasses import dataclass, field
from typing import Dict, Optional

from uniflow import Context, GuidedPrompt
from uniflow.op.extract.split.constants import PARAGRAPH_SPLITTER
from uniflow.op.model.model_config import (
    HuggingfaceModelConfig,
    LMQGModelConfig,
    ModelConfig,
    NougatModelConfig,
    OpenAIModelConfig,
)

###########################################################
#                   All Extract Configs                   #
###########################################################


@dataclass
class ExtractConfig:
    """Extract Config Class."""

    flow_name: str
    num_thread: int = 1
    model_config: Optional[ModelConfig] = None
    splitter: Optional[str] = None


@dataclass
class ExtractTxtConfig(ExtractConfig):
    """Extract Txt Config Class."""

    flow_name: str = "ExtractTxtFlow"


@dataclass
class ExtractPDFConfig(ExtractConfig):
    """Nougat Config Class."""

    flow_name: str = "ExtractPDFFlow"
    model_config: ModelConfig = NougatModelConfig()
    splitter: str = PARAGRAPH_SPLITTER


@dataclass
class ExtractMarkdownConfig(ExtractConfig):
    """Extract Markdown Config Class."""

    flow_name: str = "ExtractMarkdownFlow"


@dataclass
class ExtractIpynbConfig(ExtractConfig):
    """Extract ipynb Config Class."""

    flow_name: str = "ExtractIpynbFlow"


###########################################################
#                   All Transform Configs                 #
###########################################################


@dataclass
class TransformConfig:
    """Transform Base Config Class."""

    flow_name: str
    model_config: ModelConfig
    num_thread: int = 1
    guided_prompt_template: GuidedPrompt = GuidedPrompt(
        instruction="""Generate one question and its corresponding answer based on the last context in the last
    example. Follow the format of the examples below to include context, question, and answer in the response""",
        examples=[
            Context(
                context="The quick brown fox jumps over the lazy black dog.",
                question="What is the color of the fox?",
                answer="brown.",
            ),
            Context(
                context="The quick brown fox jumps over the lazy black dog.",
                question="What is the color of the dog?",
                answer="black.",
            ),
        ],
    )


@dataclass
class TransformOpenAIConfig(TransformConfig):
    """Transform OpenAI Config Class."""

    flow_name: str = "TransformOpenAIFlow"
    model_config: ModelConfig = OpenAIModelConfig()


@dataclass
class TransformHuggingFaceConfig(TransformConfig):
    """Transform Hugging Face Config Class."""

    flow_name: str = "TransformHuggingFaceFlow"
    model_config: ModelConfig = HuggingfaceModelConfig()


@dataclass
class TransformLMQGConfig(TransformConfig):
    """Transform LMQG Config Class."""

    flow_name: str = "TransformLMQGFlow"
    guided_prompt_template: GuidedPrompt = GuidedPrompt(instruction="", examples=[])
    model_config: ModelConfig = LMQGModelConfig()


@dataclass
class TransformCopyConfig(TransformConfig):
    """Transform Linear Config Class."""

    flow_name: str = "TransformCopyFlow"
    guided_prompt_template: GuidedPrompt = GuidedPrompt(instruction="", examples=[])
    model_config: ModelConfig = field(default_factory=lambda: {})


###########################################################
#                   All AutoRater Config                  #
###########################################################
@dataclass
class RaterConfig:
    """Rater Config Class."""

    flow_name: str
    model_config: ModelConfig
    label2score: Dict[str, float]
    guided_prompt_template: GuidedPrompt
    num_thread: int = 1

    def __post_init__(self):
        """Post-initialization to perform label check."""
        for example in self.guided_prompt_template.examples:
            if example.label.lower() not in [k.lower() for k in self.label2score]:
                raise ValueError(
                    "Inconsistent labels found in guided_prompt_template examples, "
                    f"example label {example.label} not in label2score has keys {list(self.label2score.keys())}",
                )

    def check_labels_in_label2score(self) -> bool:
        """
        Check if every label in the guided_prompt_template's examples is a key in label2score.

        Returns:
            bool: True if all labels are in label2score, False otherwise.
        """
        for example in self.guided_prompt_template.examples:
            if example.label not in self.label2score:
                return False
        return True


@dataclass
class RaterClassificationConfig(RaterConfig):
    """Rater classification Config Class."""

    flow_name: str = "RaterClassificationFlow"
    model_config: ModelConfig = OpenAIModelConfig()
    label2score: Dict[str, float] = field(
        default_factory=lambda: {"Yes": 1.0, "No": 0.0}
    )
    guided_prompt_template: GuidedPrompt = GuidedPrompt(
        instruction="""Rate the answer based on the question and the context.
        Follow the format of the examples below to include context, question, answer, and label in the response.
        The response should not include examples in the prompt.""",
        examples=[
            Context(
                context="The Eiffel Tower, located in Paris, France, is one of the most famous landmarks in the world. It was constructed in 1889 and stands at a height of 324 meters.",
                question="When was the Eiffel Tower constructed?",
                answer="The Eiffel Tower was constructed in 1889.",
                explanation="The context explicitly mentions that the Eiffel Tower was constructed in 1889, so the answer is correct.",
                label="Yes",
            ),
            Context(
                context="Photosynthesis is a process used by plants to convert light energy into chemical energy. This process primarily occurs in the chloroplasts of plant cells.",
                question="Where does photosynthesis primarily occur in plant cells?",
                answer="Photosynthesis primarily occurs in the mitochondria of plant cells.",
                explanation="The context mentions that photosynthesis primarily occurs in the chloroplasts of plant cells, so the answer is incorrect.",
                label="No",
            ),
        ],
    )


###########################################################
#                    Pipeline Config                      #
###########################################################
@dataclass
class PipelineConfig:
    """MultiFlowsPipeline Config Class."""

    extract_config: ExtractConfig
    transform_config: TransformConfig



"""Extract txt flow."""

from typing import Sequence

from uniflow.constants import EXTRACT
from uniflow.flow.flow import Flow
from uniflow.node import Node
from uniflow.op.extract.load.txt_op import ExtractTxtOp, ProcessTxtOp


class ExtractTxtFlow(Flow):
    """Extract txt Flow Class."""

    TAG = EXTRACT

    def __init__(self) -> None:
        """Extract txt Flow Constructor."""
        super().__init__()
        self._extract_txt_op = ExtractTxtOp(name="extract_txt_op")
        self._process_txt_op = ProcessTxtOp(name="process_txt_op")

    def run(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run Extract txt Flow.

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        nodes = self._extract_txt_op(nodes)
        nodes = self._process_txt_op(nodes)
        return nodes



"""Queue-Based Pipeline for flow streaming with multiple threads."""

from queue import Empty, Queue
from threading import Thread
from typing import Any, List, Mapping

from uniflow.flow.client import ExtractClient, TransformClient
from uniflow.flow.config import PipelineConfig
from uniflow.op.prompt_schema import Context


class MultiFlowsPipeline:
    """
    A queue-based pipeline to handle multiple processes for Uniflow,
    including a consumer thread and a producer thread. The Producer/Consumer
    pattern decouples processes that produce and consume data at different
    rates.

    The Producer thread is responsible for putting items into the queue if
    it is not full, while the Consumer thread consumes items if there are any.

    Both threads run indefinitely while checking the status of the queue.
    The Producer thread will stop once it has finished putting all the items
    into the queue. The Consumer thread will stop once it has consumed all
    the items from the queue.
    """

    def __init__(self, config: PipelineConfig) -> None:
        """MultiFlowsPipeline constructor

        Args:
            config (Dict[str, Any]): Config for the pipeline
        """
        self._queue = Queue()
        self._config = config
        self._extract_client = ExtractClient(self._config.extract_config)
        self._transform_client = TransformClient(self._config.transform_config)

    def _producer(self, input_list: List[Mapping[str, Any]]) -> None:
        """Producer function to put items into the queue."""
        for item in input_list:
            extracted_data = self._extract_client.run([item])
            self._queue.put(extracted_data)
        self._queue.put(None)  # Signal for the consumer that production is done

    def _consumer(self, output_list: List[Mapping[str, Any]]) -> None:
        """Consumer function to process items from the queue."""
        while True:
            try:
                data = self._queue.get(timeout=10)  # Adjust timeout as needed
                if data is None:
                    break  # End signal received
                reformat_data = []
                for d in data:
                    # rekey data and reformat to list format
                    rd = [Context(context=t) for t in d["output"][0]["text"]]
                    reformat_data.extend(rd)
                transformed_data = self._transform_client.run(reformat_data)
                output_list.append(transformed_data)
            except Empty:
                # Handle empty queue situation if needed
                pass

    def run(self, input_list: List[Mapping[str, Any]]) -> List[Mapping[str, Any]]:
        """Run the pipeline with producer-consumer pattern.

        Args:
            input_list (List[Mapping[str, Any]]): List of inputs to the pipeline

        Returns:
            List[Mapping[str, Any]]: List of outputs from the pipeline
        """
        output_list = []
        producer_thread = Thread(target=self._producer, args=(input_list,))
        consumer_thread = Thread(target=self._consumer, args=(output_list,))

        producer_thread.start()
        consumer_thread.start()

        producer_thread.join()
        consumer_thread.join()

        return output_list



"""Schema for uniflow."""

import logging
from typing import Any, Dict

from pydantic import BaseModel, ConfigDict, Field, conlist

logger = logging.getLogger(__name__)

python_to_json_types = {
    "dict": "object",
    "list": "array",
    "tuple": "array",
    "str": "string",
    "int": "number",
    "float": "number",
    "bool": "boolean",
    "None": "null",
}


class Context(BaseModel):
    """Type for prompt context."""

    context: str = Field(..., min_length=1)

    model_config = ConfigDict(extra="allow")

    def get_custom_schema(self) -> Dict[str, Any]:
        """Generate custom schema for the model.

        Args:
            None

        Returns:
            Dict[str, Any]: Custom schema for the model.
        """
        # Extract the known fields
        known_fields = self.model_json_schema()["properties"]

        schema_dict = self.model_dump()

        # Get keys that are not part of known_fields
        extra_properties = [key for key in schema_dict if key not in known_fields]

        schema = self.model_json_schema()

        # If there are extra properties, add them to the schema
        for prop in extra_properties:
            schema["properties"][prop] = {
                "type": python_to_json_types[type(schema_dict[prop]).__name__],
            }

        return schema["properties"]


class GuidedPrompt(BaseModel):
    """Type for guided prompt."""

    instruction: str = Field(..., min_length=0)

    examples: conlist(Context, min_length=0)

    model_config = ConfigDict(extra="forbid")



"""Flow Factory Module."""
from typing import Dict, List

from uniflow.constants import EXTRACT, RATER, TRANSFORM


class FlowFactory:
    """Flow Factory Class."""

    _flows = {EXTRACT: {}, TRANSFORM: {}, RATER: {}}

    @classmethod
    def register(cls, name: str, flow_cls: "Flow") -> None:  # noqa: F821
        """Register flow.

        Args:
            name (str): Flow name.
            flow_cls (Flow): Flow class.
        """
        if hasattr(flow_cls, "TAG") and flow_cls.TAG in cls._flows:
            cls._flows[flow_cls.TAG][name] = flow_cls

    @classmethod
    def get(cls, name: str, flow_type: str) -> "Flow":  # noqa: F821
        """Get flow.

        Args:
            name (str): Flow name.

        Returns:
            Flow: Flow.

        Raises:
            ValueError: If no flow registered under the name.
        """
        flow_cls = cls._flows.get(flow_type).get(name)
        if not flow_cls:
            raise ValueError(f"No flow registered under '{name}'")
        return flow_cls

    @classmethod
    def list(cls) -> Dict[str, List[str]]:
        """List all registered flows.

        Returns:
            List[str]: List of registered flow names.
        """
        return {
            subclass: list(subclass_dict.keys())
            for subclass, subclass_dict in cls._flows.items()
        }



"""Model config module."""

from dataclasses import dataclass, field
from typing import Dict


@dataclass
class ModelConfig:
    """Model Config Class."""

    model_name: str = "gpt-3.5-turbo-1106"
    model_server: str = "OpenAIModelServer"


@dataclass
class OpenAIModelConfig(ModelConfig):
    """OpenAI Model Config Class."""

    model_server: str = "OpenAIModelServer"
    num_call: int = 1
    temperature: float = 0.9
    response_format: Dict[str, str] = field(default_factory=lambda: {"type": "text"})


@dataclass
class AzureOpenAIModelConfig:
    """Azure OpenAI Model Config Class."""

    api_key: str
    api_version: str
    azure_endpoint: str
    model_name: str = "gpt-3.5-turbo-1106"
    model_server: str = "AzureOpenAIModelServer"
    num_call: int = 1
    temperature: float = 0.9
    response_format: Dict[str, str] = field(default_factory=lambda: {"type": "text"})


@dataclass
class HuggingfaceModelConfig(ModelConfig):
    """Huggingface Model Config Class."""

    model_name: str = "mistralai/Mistral-7B-Instruct-v0.1"
    batch_size: int = 1
    model_server: str = "HuggingfaceModelServer"


@dataclass
class LMQGModelConfig(ModelConfig):
    """LMQG Model Config Class."""

    model_name: str = "lmqg/t5-base-squad-qg-ae"
    batch_size: int = 1
    model_server: str = "LMQGModelServer"


@dataclass
class NougatModelConfig(ModelConfig):
    """Nougat Model Config Class."""

    model_name: str = "0.1.0-small"
    batch_size: int = 1
    model_server: str = "NougatModelServer"
